import logging
import functools
import operator

l = logging.getLogger("claripy.backends.backend_vsa")

from ..backend import Backend, BackendError

def arg_filter(f):
    @functools.wraps(f)
    def filter(*args): #pylint:disable=redefined-builtin
        if type(args[0]) in {int, long}: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported argument type %s' % type(args[0]))
        return f(*args)

    return filter

def normalize_arg_order(f):
    @functools.wraps(f)
    def normalizer(*args):
        if len(args) != 2:
            raise BackendError('Unsupported arguments number %d' % len(args))

        if type(args[0]) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            if type(args[1]) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
                raise BackendError('Unsupported arguments')
            args = [args[1], args[0]]

        return f(*args)

    return normalizer

def normalize_reversed_arguments(f):
    @functools.wraps(f)
    def normalizer(self, ast, result=None):
        arg_reversed = []
        raw_args = []
        for i in xrange(len(ast.args)):
            if isinstance(ast.args[i], Base) and \
                            type(self.convert(ast.args[i])) in { #pylint:disable=unidiomatic-typecheck
                                                    StridedInterval,
                                                    DiscreteStridedIntervalSet,
                                                    ValueSet
            }:
                if self.convert(ast.args[i]).reversed:
                    arg_reversed.append(True)
                    raw_args.append(ast.args[i].reversed)
                    continue

            # It's not reversed
            arg_reversed.append(False)
            raw_args.append(ast.args[i])

        any_reversed_arg = any(arg_reversed)
        for i in xrange(len(raw_args)):
            raw_args[i] = self.convert(raw_args[i])

        normalized = ast.swap_args(raw_args)
        ret = f(self, normalized, result=result)

        # inner_i = I(args[0]._claripy, ret, variables=variables)
        if any_reversed_arg:
            return ret.reverse()
            #ret = A(args[0]._claripy, 'Reverse', (inner_i,), variables=variables, collapsible=False)

        return ret

    return normalizer

class BackendVSA(Backend):
    def __init__(self):
        Backend.__init__(self)
        # self._make_raw_ops(set(expression_operations) - set(expression_set_operations), op_module=BackendVSA)
        self._make_expr_ops(set(expression_set_operations), op_class=self)
        self._make_raw_ops(set(backend_operations_vsa_compliant), op_module=BackendVSA)

        self._op_raw['StridedInterval'] = BackendVSA.CreateStridedInterval
        self._op_raw['ValueSet'] = ValueSet.__init__
        self._op_raw['AbstractLocation'] = AbstractLocation.__init__
        self._op_raw['Reverse'] = BackendVSA.Reverse
        self._op_raw['If'] = self.If
        self._op_expr['BVV'] = self.BVV
        self._op_expr['BoolV'] = self.BoolV
        self._op_expr['BVS'] = self.BVS

        self._balancer = Balancer(self)

    def convert(self, expr, result=None):
        return Backend.convert(self, expr.ite_excavated if isinstance(expr, Base) else expr, result=result)

    def _convert(self, a, result=None):
        if type(a) in { int, long }: #pylint:disable=unidiomatic-typecheck
            return a
        if type(a) is bool:
            return TrueResult() if a else FalseResult()
        if type(a) in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            return a
        if isinstance(a, BoolResult):
            return a

        raise BackendError("why is fish raising NotImplementedError INSTEAD OF THE ERROR THAT'S SUPPOSED TO BE RAISED IN THIS SITUATION? SERIOUSLY, JUST RAISE A BACKENDERROR AND EVERYONE WILL BE HAPPY, BUT NO, PEOPLE HAVE TO RAISE THEIR OWN ERRORS INSTEAD OF USING THE ERRORS THAT WERE ****DESIGNED**** FOR THIS SORT OF THING. WHY DO I BOTHER DESIGNING A GOOD ERROR HIERARCHY, ANYWAYS? WILL IT BE USED? NO! IT'LL BE ALL NotImplementedError('THIS') or Exception('THAT') AND EVERYTHING WILL MELT DOWN. UGH!")

    def _eval(self, expr, n, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, StridedInterval):
            return expr.eval(n)
        elif isinstance(expr, ValueSet):
            results = []

            while len(results) < n:
                results.extend(expr.eval(n - len(results)))

            return results
        elif isinstance(expr, BoolResult):
            return expr.value
        else:
            raise BackendError('Unsupported type %s' % type(expr))

    def _min(self, expr, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, StridedInterval):
            if expr.is_top:
                # TODO: Return
                return 0

            return expr.min
        else:
            raise BackendError('Unsupported expr type %s' % type(expr))

    def _max(self, expr, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, StridedInterval):
            if expr.is_top:
                # TODO:
                return StridedInterval.max_int(expr.bits)

            return expr.max

        else:
            raise BackendError('Unsupported expr type %s' % type(expr))

    def _solution(self, obj, v, result=None, solver=None, extra_constraints=()):
        if isinstance(obj, BoolResult):
            return len(set(v.value) & set(obj.value)) > 0

        if isinstance(obj, StridedInterval):
            return not obj.intersection(v).is_empty

        if isinstance(obj, ValueSet):
            for _, si in obj.items():
                if not si.intersection(v).is_empty:
                    return True
            return False

        raise NotImplementedError(type(obj).__name__)

    def _has_true(self, o, extra_constraints=(), result=None, solver=None):
        return BoolResult.has_true(o)

    def _has_false(self, o, extra_constraints=(), result=None, solver=None):
        return BoolResult.has_false(o)

    def _is_true(self, o, extra_constraints=(), result=None, solver=None):
        return BoolResult.is_true(o)

    def _is_false(self, o, extra_constraints=(), result=None, solver=None):
        return BoolResult.is_false(o)

    #
    # Backend Operations
    #

    def simplify(self, e):
        raise BackendError('nope')

    def _identical(self, a, b, result=None):
        if type(a) != type(b):
            return False
        return a.identical(b)

    def _unique(self, obj, result=None): #pylint:disable=unused-argument,no-self-use
        if isinstance(obj, StridedInterval):
            return obj.unique
        elif isinstance(obj, ValueSet):
            return obj.unique
        else:
            raise BackendError('Not supported type of operand %s' % type(obj))

    def _cardinality(self, a, result=None): #pylint:disable=unused-argument,no-self-use
        return a.cardinality

    def name(self, a, result=None):
        if isinstance(a, StridedInterval):
            return a.name

        else:
            return None

    def BVV(self, ast, result=None): #pylint:disable=unused-argument
        if ast.args[0] is None:
            return StridedInterval.empty(ast.args[1])
        else:
            return self.CreateStridedInterval(bits=ast.args[1], stride=0, lower_bound=ast.args[0], upper_bound=ast.args[0])

    @staticmethod
    def BoolV(ast, result=None): #pylint:disable=unused-argument
        return TrueResult() if ast.args[0] else FalseResult()

    @staticmethod
    def And(a, *args):
        return reduce(operator.__and__, args, a)

    @staticmethod
    def Not(a):
        return ~a

    @staticmethod
    @normalize_arg_order
    def ULT(a, b):
        return a.ULT(b)

    @staticmethod
    @normalize_arg_order
    def ULE(a, b):
        return a.ULE(b)

    @staticmethod
    @normalize_arg_order
    def UGT(a, b):
        return a.UGT(b)

    @staticmethod
    @normalize_arg_order
    def UGE(a, b):
        return a.UGE(b)

    @staticmethod
    @normalize_arg_order
    def SLT(a, b):
        return a.SLT(b)

    @staticmethod
    @normalize_arg_order
    def SLE(a, b):
        return a.SLE(b)

    @staticmethod
    @normalize_arg_order
    def SGT(a, b):
        return a.SGT(b)

    @staticmethod
    @normalize_arg_order
    def SGE(a, b):
        return a.SGE(b)

    @staticmethod
    def BVS(ast, result=None): #pylint:disable=unused-argument
        size = ast.size()
        name, mn, mx, stride, uninitialized = ast.args
        return CreateStridedInterval(name=name, bits=size, lower_bound=mn, upper_bound=mx, stride=stride, uninitialized=uninitialized)

    def If(self, cond, t, f):
        if not self.has_true(cond):
            return f
        elif not self.has_false(cond):
            return t
        else:
            return t.union(f)

    # TODO: Implement other operations!

    @staticmethod
    def LShR(expr, shift_amount):
        return expr >> shift_amount

    @staticmethod
    def Concat(*args):
        ret = None
        for expr in args:
            if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
                raise BackendError('Unsupported expr type %s' % type(expr))

            ret = ret.concat(expr) if ret is not None else expr

        return ret

    @arg_filter
    def _size(self, arg, result=None):
        if type(arg) in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            return len(arg)
        else:
            return arg.size()

    @staticmethod
    def Extract(*args):
        low_bit = args[1]
        high_bit = args[0]
        expr = args[2]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        ret = expr.extract(high_bit, low_bit)

        return ret

    @staticmethod
    def SignExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        return expr.sign_extend(new_bits + expr.bits)

    @staticmethod
    def ZeroExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        return expr.zero_extend(new_bits + expr.bits)

    @staticmethod
    def Reverse(arg):
        if type(arg) not in {StridedInterval, DiscreteStridedIntervalSet, ValueSet}: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(arg))

        return arg.reverse()

    @normalize_reversed_arguments
    def union(self, ast, result=None): #pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.union().' % len(ast.args))

        ret = ast.args[0].union(ast.args[1])

        if ret is NotImplemented:
            ret = ast.args[1].union(ast.args[0])

        return ret

    @normalize_reversed_arguments
    def intersection(self, ast, result=None): #pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.intersection().' % len(ast.args))

        ret = None

        for arg in ast.args:
            if ret is None:
                ret = arg
            else:
                ret = ret.intersection(arg)

        return ret

    @normalize_reversed_arguments
    def widen(self, ast, result=None): #pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.widen().' % len(ast.args))

        ret = ast.args[0].widen(ast.args[1])
        if ret is NotImplemented:
            ret = ast.args[1].widen(ast.args[0])

        return ret

    @staticmethod
    def CreateTopStridedInterval(bits, name=None, uninitialized=False): #pylint:disable=unused-argument,no-self-use
        return StridedInterval.top(bits, name, uninitialized=uninitialized)

    def constraint_to_si(self, expr):
        return self._balancer.constraint_to_si(expr)

from ..ast.base import Base
from ..operations import backend_operations_vsa_compliant, expression_set_operations
from ..vsa import StridedInterval, CreateStridedInterval, DiscreteStridedIntervalSet, ValueSet, AbstractLocation, BoolResult, TrueResult, FalseResult
from ..balancer import Balancer

BackendVSA.CreateStridedInterval = staticmethod(CreateStridedInterval)



import logging
l = logging.getLogger("claripy.ops")

#
# AST creation
#

def AbstractLocation(*args, **kwargs): #pylint:disable=no-self-use
    aloc = vsa.AbstractLocation(*args, **kwargs)
    return aloc

#
# Some operations
#


#
# sigh
#

#pylint:disable=wildcard-import,unused-wildcard-import
from .ast.base import *
from .ast.bv import *
from .ast.fp import *
from .ast.bool import *
from . import vsa

VS = ValueSet
#SI = StridedInterval
#TSI = TopStridedInterval
#ESI = EmptyStridedInterval



#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.hybrid_frontend")

from .full_frontend import FullFrontend
from .light_frontend import LightFrontend
from .replacement_frontend import ReplacementFrontend

class HybridFrontend(FullFrontend):
    def __init__(self, solver_backend, approximation_frontend=None, **kwargs):
        FullFrontend.__init__(self, solver_backend, **kwargs)
        self._approximation_frontend = approximation_frontend if approximation_frontend is not None else ReplacementFrontend(LightFrontend(backends.vsa), replace_constraints=True)

    #
    # Storable support
    #

    def _ana_getstate(self):
        return (self._approximation_frontend, FullFrontend._ana_getstate(self))

    def _ana_setstate(self, s):
        self._approximation_frontend, full_state = s
        FullFrontend._ana_setstate(self, full_state)

    def _blank_copy(self):
        return HybridFrontend(self._solver_backend, approximation_frontend=self._approximation_frontend._blank_copy(), cache=self._cache, timeout=self.timeout)

    def branch(self):
        s = super(HybridFrontend, self).branch()
        s._approximation_frontend = self._approximation_frontend.branch()
        return s

    #
    # Catch constraint adds and pass them to the replacement frontend
    #

    def _add_constraints(self, constraints, invalidate_cache=True):
        super(HybridFrontend, self)._add_constraints(constraints, invalidate_cache=invalidate_cache)
        self._approximation_frontend.add(constraints, invalidate_cache=invalidate_cache)

    def _cache_eval(self, e, values, n=None, exact=None, cache=None):
        super(HybridFrontend, self)._cache_eval(e, values, n=None, exact=exact, cache=cache)

        if not exact is False:
            if n > 1 and len(values) == 1:
                self._approximation_frontend.add_replacement(e, next(iter(values)))

    def _cache_max(self, e, mx, exact=None, cache=None):
        super(HybridFrontend, self)._cache_max(e, mx, exact=exact, cache=cache)

        if not exact is False and not cache is False:
            if isinstance(e, BV):
                try:
                    if self.result and e.uuid in self.result.min_cache:
                        mn = self.result.min_cache[e.uuid]
                    else:
                        mn = self._approximation_frontend.min(e)
                except ClaripyFrontendError:
                    mn = 0

                si = BVS('limiter', e.length, max=mx, min=mn)
                self._approximation_frontend.add_replacement(e, e.intersection(si))

    def _cache_min(self, e, mn, exact=None, cache=None):
        super(HybridFrontend, self)._cache_min(e, mn, exact=exact, cache=cache)

        if not exact is False and not cache is False:
            if isinstance(e, BV):
                try:
                    if self.result and e.uuid in self.result.max_cache:
                        mx = self.result.max_cache[e.uuid]
                    else:
                        mx = self._approximation_frontend.max(e)
                except ClaripyFrontendError:
                    mx = 0

                si = BVS('limiter', e.length, min=mn, max=mx)
                self._approximation_frontend.add_replacement(e, e.intersection(si))

    #
    # Hybrid solving
    #

    def _solve(self, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.solve(extra_constraints=extra_constraints)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _solve()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._solve(extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.eval(e, n, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _eval()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._eval(e, n, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):

        if exact is False:
            try:
                return self._approximation_frontend.batch_eval(exprs, n, extra_constraints=extra_constraints)

            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _batch_eval()")

        return super(HybridFrontend, self)._batch_eval(
                exprs,
                n,
                extra_constraints=extra_constraints,
                exact=exact,
                cache=cache
        )

    def _max(self, e, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.max(e, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _max()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._max(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _min(self, e, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.min(e, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _min()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._min(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.solution(e, v, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _solution()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._solution(e, v, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _is_true(self, e, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.is_true(e, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _is_true()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._is_true(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _is_false(self, e, extra_constraints=(), exact=None, cache=None):
        # if approximating, try the approximation backend
        if exact is False:
            try:
                return self._approximation_frontend.is_false(e, extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _is_false()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._is_false(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

    def _satisfiable(self, extra_constraints=(), exact=None, cache=None):
        if exact is False:
            try:
                new_constraints = self._approximation_frontend._replace_list(self._approximation_frontend.constraints)
                self._approximation_frontend.constraints[:] = new_constraints
                return self._approximation_frontend.satisfiable(extra_constraints=extra_constraints, cache=cache)
            except ClaripyFrontendError:
                l.debug("Approximation failed. Falling back on exact _satisfiable()")

        # if that fails, try the exact backend
        return super(HybridFrontend, self)._satisfiable(extra_constraints=extra_constraints, exact=exact, cache=cache)


from ..errors import ClaripyFrontendError
from ..backend_manager import backends
from ..ast.bv import BV, BVS



Bits = lambda *args, **kwargs: None
BV = lambda *args, **kwargs: None
VS = lambda *args, **kwargs: None
FP = lambda *args, **kwargs: None
Bool = lambda *args, **kwargs: None
Int = lambda *args, **kwargs: None
Base = lambda *args, **kwargs: None
all_operations = None

def _import():
    global Bits, BV, VS, FP, Bool, Int, Base, all_operations

    from .bits import Bits
    from .bv import BV
    from .vs import VS
    from .fp import FP
    from .bool import Bool
    from .int import Int
    from .base import Base
    from .. import ops as all_operations



#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.frontend")

import ana
import sys

#pylint:disable=unidiomatic-typecheck

class Frontend(ana.Storable):
    def __init__(self, cache=None):
        self.result = None
        self._simplified = False
        self._cache = cache is not False

    #
    # Storable support
    #

    @property
    def uuid(self):
        return self.ana_uuid

    def _ana_getstate(self):
        if not self._simplified: self.simplify()
        return self.result

    def _ana_setstate(self, s):
        self.result = s
        self._simplified = True

    def branch(self):
        s = self._blank_copy()
        s.result = self.result
        s._simplified = self._simplified
        return s

    def _blank_copy(self):
        return Frontend(cache=self._cache)

    #
    # Constraint management
    #

    @staticmethod
    def _split_constraints(constraints):
        '''
        Returns independent constraints, split from this Frontend's constraints.
        '''

        splitted = [ ]
        for i in constraints:
            splitted.extend(i.split(['And']))

        l.debug("... splitted of size %d", len(splitted))

        variable_connections = { }
        constraint_connections = { }
        for n,s in enumerate(splitted):
            l.debug("... processing constraint with %d variables", len(s.variables))

            connected_variables = set(s.variables)
            connected_constraints = { n }

            if len(connected_variables) == 0:
                connected_variables.add('CONCRETE')

            for v in s.variables:
                if v in variable_connections:
                    connected_variables |= variable_connections[v]
                if v in constraint_connections:
                    connected_constraints |= constraint_connections[v]

            for v in connected_variables:
                variable_connections[v] = connected_variables
                constraint_connections[v] = connected_constraints

        unique_constraint_sets = set()
        for v in variable_connections:
            unique_constraint_sets.add((frozenset(variable_connections[v]), frozenset(constraint_connections[v])))

        results = [ ]
        for v,c_indexes in unique_constraint_sets:
            results.append((set(v), [ splitted[c] for c in c_indexes ]))
        return results

    def _filter_single_constraint(self, e_simp): #pylint:disable=no-self-use
        if not isinstance(e_simp, (Bool, bool)):
            l.warning("Frontend._constraint_filter got non-boolean from model_backend")
            raise ClaripyFrontendError()

        if self._eager_resolution('is_false', False, e_simp, use_result=False):
            raise UnsatError("expressions contain False")
        elif self._eager_resolution('is_true', False, e_simp, use_result=False):
            return None
        else:
            return e_simp

    def _constraint_filter(self, ec):
        fc = [ ]
        for e in ec if type(ec) in (list, tuple, set) else (ec,):
            #e_simp = self._claripy.simplify(e)
            c = self._filter_single_constraint(e)
            if c is not None:
                fc.append(c)

        return tuple(fc)

    #
    # Stuff that should be implemented by subclasses
    #

    def _add_constraints(self, constraints, invalidate_cache=True):
        raise NotImplementedError("_add_constraints() is not implemented")

    def _simplify(self):
        raise NotImplementedError("_simplify() is not implemented")

    def _solve(self, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_solve() is not implemented")

    def _eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_eval() is not implemented")

    def _batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_batch_eval() is not implemented")

    def _min(self, e, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_min() is not implemented")

    def _max(self, e, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_max() is not implemented")

    def _solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_solution() is not implemented")

    def _is_true(self, e, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_is_true() is not implemented")

    def _is_false(self, e, extra_constraints=(), exact=None, cache=None):
        raise NotImplementedError("_is_false() is not implemented")

    def finalize(self):
        raise NotImplementedError("finalize() is not implemented")

    def merge(self, others, merge_flag, merge_values):
        raise NotImplementedError("merge() is not implemented")

    def combine(self, others):
        raise NotImplementedError("combine() is not implemented")

    def split(self):
        raise NotImplementedError("split() is not implemented")

    #
    # Caching
    #

    def _cache_eval(self, e, values, n=None, exact=None, cache=None):
        if exact is False or cache is False or not self._cache or self.result is None:
            return

        self.result.eval_cache[e.uuid] = self.result.eval_cache[e.uuid] | values if e.uuid in self.result.eval_cache else values
        if n is not None:
            self.result.eval_n[e.uuid] = max(n, self.result.eval_n[e.uuid]) if e.uuid in self.result.eval_n else n

    def _cache_max(self, e, m, exact=None, cache=None):
        if exact is False or cache is False or not self._cache or self.result is None:
            return

        self._cache_eval(e, {m}, exact=exact, cache=cache)
        self.result.max_cache[e.uuid] = min(self.result.max_cache.get(e, (2**e.length)-1), m)

    def _cache_min(self, e, m, exact=None, cache=None):
        if exact is False or cache is False or not self._cache or self.result is None:
            return

        self._cache_eval(e, {m}, exact=exact, cache=cache)
        self.result.min_cache[e.uuid] = max(self.result.min_cache.get(e, 0), m)

    #
    # Solving
    #

    def add(self, constraints, invalidate_cache=True):
        if type(constraints) not in (list, tuple):
            constraints = [ constraints ]

        if len(constraints) == 0:
            return [ ]

        try:
            to_add = self._constraint_filter(constraints)
        except UnsatError:
            self.result = UnsatResult()
            to_add = [ false ]

        for c in to_add:
            if not isinstance(c, Bool):
                raise ClaripyTypeError('constraint is not a boolean expression!')
            c.make_uuid()

        if self.result is not None and invalidate_cache:
            all_true = True
            for c in to_add:
                v = self._eager_resolution('eval', [None], c, 1)[0]
                if v is None:
                    all_true = False
                    break
                else:
                    all_true &= v
        else:
            all_true = False

        self._add_constraints(to_add, invalidate_cache=invalidate_cache)
        self._simplified = False

        if invalidate_cache and self.result is not None and self.result.sat:
            if all_true:
                new_result = SatResult()
                new_result.model.update(self.result.model)
                self.result = new_result
            else:
                self.result = None

        return to_add

    def simplify(self):
        if self._simplified:
            return

        s = self._simplify()
        self._simplified = True
        return s

    def solve(self, extra_constraints=(), exact=None, cache=None):
        l.debug("%s.solve() running with %d extra constraints...", self.__class__.__name__, len(extra_constraints))

        if self.result is not None:
            if not self.result.sat or len(extra_constraints) == 0:
                l.debug("... returning cached result (sat: %s)", self.result.sat)
                return self.result
        else:
            l.debug("... no cached result")

        try:
            extra_constraints = self._constraint_filter(extra_constraints)
        except UnsatError:
            l.debug("... returning unsat result due to false extra_constraints")
            return UnsatResult()

        l.debug("... conferring with the solver")
        r = self._solve(extra_constraints=extra_constraints, exact=exact, cache=cache)
        if exact is not False and cache is not False and (len(extra_constraints) == 0 or (self.result is None and r.sat)):
            l.debug("... caching result (sat: %s)", r.sat)
            self.result = r
        return r

    def satisfiable(self, extra_constraints=(), exact=None, cache=None):
        return self.solve(extra_constraints=extra_constraints, exact=exact, cache=cache).sat

    @staticmethod
    def _concrete_type_check(e):
        '''
        Checks two things:

            1. Whether we can just return this value.
            2. Whether we can even process this value.

        Returns True if we don't have to pass this to any backends, False if we need
        to, and raises ClaripyValueError otherwise.
        '''

        if isinstance(e, (int, long)):
            return True
        elif not isinstance(e, Base):
            raise ClaripyValueError("Expressions passed to min() MUST be Claripy ASTs (got %s)" % type(e))
        else:
            return False

    def eval_to_ast(self, e, n, extra_constraints=(), exact=None, cache=None):
        '''
        Evaluates expression e, returning the results in the form of concrete ASTs.
        '''

        return [ BVV(v, e.size()) for v in self.eval(e, n, extra_constraints=extra_constraints, exact=exact, cache=cache) ]

    def _eager_resolution(self, what, default, *args, **kwargs):
        for b in backends._eager_backends:
            try: return getattr(b, what)(*args, result=self.result if kwargs.pop('use_result', True) else None, **kwargs)
            except BackendError: pass
        return default

    def eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        '''
        Evaluates an expression, returning a tuple of n solutions.

        @param e: the expression
        @param n: the number of solutions to return
        @param extra_constraints: extra constraints to consider when performing
                                  the evaluation
        @param exact: whether or not to perform an exact evaluation. Ignored by
                      non-approximating backends.
        @param cache: whether or not to cache the results

        @returns a tuple of python pritives representing results
        '''

        if self._concrete_type_check(e): return (e,)

        extra_constraints = self._constraint_filter(extra_constraints)
        l.debug("Frontend.eval() for UUID %s with n=%d and %d extra_constraints", e.uuid, n, len(extra_constraints))

        # first, try evaluating through the eager backends
        try:
            eager_results = frozenset(self._eager_resolution('eval', (), e, n, extra_constraints=extra_constraints))
            if not e.symbolic and len(eager_results) > 0:
                return tuple(sorted(eager_results))
            self._cache_eval(e, eager_results, exact=exact, cache=cache)
        except UnsatError:
            # this can happen when the eager backend comes across an unsat extra condition
            # *while using the current model*. A new constraint solve could return a new,
            # sat model
            pass

        # then, check the cache
        if len(extra_constraints) == 0 and self.result is not None and e.uuid in self.result.eval_cache:
            cached_results = self.result.eval_cache[e.uuid]
            cached_n = self.result.eval_n.get(e.uuid, 0)
        else:
            cached_results = frozenset()
            cached_n = 0

        # if there's enough in the cache, return that
        if cached_n >= n or len(cached_results) < cached_n or len(cached_results) >= n:
            return tuple(sorted(cached_results))[:n]

        # try to make sure we don't get more of the same
        solver_extra_constraints = list(extra_constraints) + [e != v for v in cached_results]

        # if we still need more results, get them from the frontend
        try:
            n_lacking = n - len(cached_results)
            eval_results = frozenset(
                    self._eval(e, n_lacking, extra_constraints=solver_extra_constraints, exact=exact, cache=cache)
            )
            l.debug("... got %d more values", len(eval_results))
        except UnsatError:
            l.debug("... UNSAT")
            if len(cached_results) == 0:
                raise
            else:
                eval_results = frozenset()
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during eval: %s('%s')" % (str(e_type), str(value)), traceback

        # if, for some reason, we have no result object, make an approximate one
        if self.result is None:
            self.result = SatResult(approximation=True)

        # if there are less possible solutions than n (i.e., meaning we got all the solutions for e),
        # add constraints to help the solver out later
        # TODO: does this really help?
        all_results = cached_results | eval_results
        if len(extra_constraints) == 0 and len(all_results) < n:
            l.debug("... adding constraints for %d values for future speedup", len(all_results))
            self.add([Or(*[e == v for v in eval_results | cached_results])], invalidate_cache=False)

        # fix up the cache. If there were extra constraints, we can't assume that we got
        # all of the possible solutions, so we have to settle for a biggest-evaluated value
        # equal to the number of values we got
        self._cache_eval(e, all_results, n=n if len(extra_constraints) == 0 else None, exact=exact, cache=cache)

        # sort so the order of results is consistent when using pypy
        return tuple(sorted(all_results))

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):

        def _formulate_results(all_exprs, symbolic_exprs, results):
            """
            Build a list of results with both concrete and symbolic expressions from a list of all expressions and
            results of symbolic expressions.

            :param all_exprs: All expressions, including both concrete and symbolic
            :param symbolic_exprs: A list of all symbolic expressions
            :param results: A list of all results for symbolic expressions
            :return: A list of tuples, where each tuple corresponds to a complete list of solutions to all expressions
            """

            # Offset of each expr expr in symbolic_exprs
            # None - concrete expr
            # int - offset in symbolic_exprs
            expr_offsets = [ ]

            symbolic_expr_pos = 0
            for expr in all_exprs:
                if symbolic_expr_pos < len(symbolic_exprs) and \
                                expr is symbolic_exprs[symbolic_expr_pos]:
                    expr_offsets.append(symbolic_expr_pos)
                    symbolic_expr_pos += 1
                else:
                    expr_offsets.append(None)

            formulated_results = [ ]
            for r in results:
                formulated_result = [ ]
                for i, offset in enumerate(expr_offsets):
                    if offset is None:
                        # concrete expr
                        formulated_result.append(all_exprs[i])
                    else:
                        # symbolic expr
                        formulated_result.append(r[offset])
                formulated_results.append(tuple(formulated_result))

            return formulated_results

        def _expr_from_results(exprs, results):
            """
            Generate a list of expressions based on the given expression list and result list to constrain the values of
            those expressions.

            :param exprs: A list of expressions
            :param results: A list of results for those expressions
            :return: A list of claripy expressions.
            """
            return [ Not(And(*[ ex == v for ex, v in zip(exprs, values) ])) for values in results ]

        symbolic_exprs = [ ]

        for ex in exprs:
            if not self._concrete_type_check(ex):
                symbolic_exprs.append(ex)

        if not symbolic_exprs:
            return [ exprs ]

        extra_constraints = self._constraint_filter(extra_constraints)
        l.debug("Frontend.batch_eval() for UUID %s with n=%d and %d extra_constraints",
                [ ex.uuid for ex in symbolic_exprs ],
                n,
                len(extra_constraints)
                )

        # first, try evaluating through the eager backends
        try:
            eager_results = frozenset(
                    self._eager_resolution('batch_eval', (), symbolic_exprs, n, extra_constraints=extra_constraints))
            if all([ not ex.symbolic for ex in symbolic_exprs ]) and len(eager_results) > 0:
                return _formulate_results(exprs, symbolic_exprs, eager_results)
        except UnsatError:
            # this can happen when the eager backend comes across an unsat extra condition
            # *while using the current model*. A new constraint solve could return a new,
            # sat model
            eager_results = [ ]

        # if there's enough from eager backends, return that
        if len(eager_results) >= n:
            return _formulate_results(exprs, symbolic_exprs, list(eager_results)[ : n])

        # try to make sure we don't get more of the same
        solver_extra_constraints = list(extra_constraints) + _expr_from_results(symbolic_exprs, eager_results)

        # if we still need more results, get them from the frontend
        try:
            n_lacking = n - len(eager_results)
            eval_results = frozenset(
                self._batch_eval(
                        symbolic_exprs, n_lacking, extra_constraints=solver_extra_constraints, exact=exact, cache=cache
                )
            )
            l.debug("... got %d more values", len(eval_results))
        except UnsatError:
            l.debug("... UNSAT")
            if len(eager_results) == 0:
                raise
            else:
                eval_results = frozenset()
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during eval: %s('%s')" % (str(e_type), str(value)), traceback

        # if, for some reason, we have no result object, make an approximate one
        if self.result is None:
            self.result = SatResult(approximation=True)

        all_results = eager_results | eval_results
        return _formulate_results(exprs, symbolic_exprs, all_results)

    def max(self, e, extra_constraints=(), exact=None, cache=None):
        if self._concrete_type_check(e): return e
        extra_constraints = self._constraint_filter(extra_constraints)

        # first, try evaluating through the eager backends
        v = self._eager_resolution('max', None, e, extra_constraints=extra_constraints, use_result=False)
        if v is not None:
            return v

        if len(extra_constraints) == 0 and self.result is not None and e.uuid in self.result.max_cache:
            #cached_max += 1
            return self.result.max_cache[e.uuid]

        m = self._max(e, extra_constraints=extra_constraints, exact=exact, cache=cache)
        if len(extra_constraints) == 0 and e.symbolic:
            self._cache_max(e, m, exact=exact, cache=cache)
            self.add([ULE(e, m)], invalidate_cache=False)
        return m

    def min(self, e, extra_constraints=(), exact=None, cache=None):
        if self._concrete_type_check(e): return e
        extra_constraints = self._constraint_filter(extra_constraints)

        # first, try evaluating through the eager backends
        v = self._eager_resolution('min', None, e, extra_constraints=extra_constraints, use_result=False)
        if v is not None:
            return v

        if len(extra_constraints) == 0 and self.result is not None and e.uuid in self.result.min_cache:
            #cached_min += 1
            return self.result.min_cache[e.uuid]

        m = self._min(e, extra_constraints=extra_constraints, exact=exact, cache=cache)
        if len(extra_constraints) == 0 and e.symbolic:
            self._cache_min(e, m, exact=exact, cache=cache)
            self.add([UGE(e, m)], invalidate_cache=False)
        return m

    def solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        try:
            extra_constraints = self._constraint_filter(extra_constraints)
        except UnsatError:
            return False

        if self._concrete_type_check(e) and self._concrete_type_check(v):
            return e == v
        eager_solution = self._eager_resolution('solution', None, e, v)
        if eager_solution is not None:
            return eager_solution

        b = self._solution(e, v, extra_constraints=extra_constraints, exact=exact, cache=cache)
        if b is False and len(extra_constraints) == 0 and e.symbolic:
            self.add([e != v], invalidate_cache=False)

        # add these results to the cache
        if self.result is not None and b is True:
            if isinstance(e, Base) and e.symbolic and not isinstance(v, Base):
                self._cache_eval(e, frozenset({v}), exact=exact, cache=cache)
            if isinstance(v, Base) and v.symbolic and not isinstance(e, Base):
                self._cache_eval(v, frozenset({e}), exact=exact, cache=cache)

        return b

    def is_true(self, e, extra_constraints=(), exact=None, cache=None):
        if self._concrete_type_check(e):
            return e is True

        if not isinstance(e, Bool):
            raise ClaripyValueError("got a non-Boolean expression in Frontend.is_true()")

        eager_solution = self._eager_resolution('is_true', None, e)
        if eager_solution is True:
            return eager_solution

        try:
            b = self._is_true(e, extra_constraints=extra_constraints, exact=exact, cache=cache)
            if b is True and len(extra_constraints) == 0 and e.symbolic:
                self.add([e == True], invalidate_cache=False)
            return b
        except ClaripyFrontendError:
            return False

    def is_false(self, e, extra_constraints=(), exact=None, cache=None):
        if self._concrete_type_check(e):
            return e is False

        if not isinstance(e, Bool):
            raise ClaripyValueError("got a non-Boolean expression in Frontend.is_false()")

        eager_solution = self._eager_resolution('is_false', None, e)
        if eager_solution is True:
            return eager_solution

        try:
            b = self._is_false(e, extra_constraints=extra_constraints, exact=exact, cache=cache)
            if b is True and len(extra_constraints) == 0 and e.symbolic:
                self.add([e == False], invalidate_cache=False)
            return b
        except ClaripyFrontendError:
            return False

    #
    # Serialization and such.
    #

    def downsize(self): #pylint:disable=R0201
        if self.result is not None:
            self.result.downsize()

from .result import UnsatResult, SatResult
from .errors import UnsatError, BackendError, ClaripyFrontendError, ClaripyTypeError, ClaripyValueError
from .backend_manager import backends
from .ast.base import Base
from .ast.bool import false, Bool, And, Or, Not
from .ast.bv import UGE, ULE, BVV



from distutils.core import setup

setup(
    name='claripy',
    version='4.6.1.18',
    packages=['claripy', 'claripy.backends', 'claripy.frontends', 'claripy.vsa', 'claripy.ast'],
    install_requires=[
        'ana',
        'angr-z3',
    ],
    description='An abstraction layer for constraint solvers',
    url='https://github.com/angr/claripy',
)



#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.full_frontend")

import sys
import threading

from .constrained_frontend import ConstrainedFrontend

class FullFrontend(ConstrainedFrontend):
    def __init__(self, solver_backend, timeout=None, **kwargs):
        ConstrainedFrontend.__init__(self, **kwargs)
        self._solver_backend = solver_backend
        self.timeout = timeout if timeout is not None else 300000
        self._tls = threading.local()
        self._to_add = [ ]

    #
    # Storable support
    #

    def _ana_getstate(self):
        return self._solver_backend.__class__.__name__, self.timeout, ConstrainedFrontend._ana_getstate(self)

    def _ana_setstate(self, s):
        backend_name, self.timeout, base_state = s
        self._solver_backend = backends._backends_by_type[backend_name]
        self._tls = None
        self._to_add = [ ]
        ConstrainedFrontend._ana_setstate(self, base_state)

    #
    # Frontend Creation
    #

    def _get_solver(self):
        if getattr(self._tls, 'solver', None) is None or (self._finalized and len(self._to_add) > 0):
            self._tls.solver = self._solver_backend.solver(timeout=self.timeout)
            self._solver_backend.add(self._tls.solver, self.constraints)
            self._to_add = [ ]

        if len(self._to_add) > 0:
            self._solver_backend.add(self._tls.solver, self._to_add)
            self._to_add = [ ]

        return self._tls.solver

    #
    # Constraint management
    #

    def _add_constraints(self, constraints, invalidate_cache=True):
        to_add = ConstrainedFrontend._add_constraints(self, constraints, invalidate_cache=invalidate_cache)
        self._to_add += to_add
        return to_add

    def _simplify(self):
        ConstrainedFrontend._simplify(self)

        # TODO: should we do this?
        self._tls.solver = None
        self._to_add = [ ]
        self._simplified = True

        return self.constraints

    def _solve(self, extra_constraints=(), exact=None, cache=None):
        r = ConstrainedFrontend._solve(self, extra_constraints=extra_constraints)
        if not r.approximation:
            return r

        l.debug("Frontend.solve() checking SATness of %d constraints", len(self.constraints))

        try:
            s = self._get_solver()

            #a = time.time()
            r = self._solver_backend.results(s, extra_constraints, generic_model=True)
            #b = time.time()
            #l_timing.debug("... %s in %s seconds", r.sat, b - a)
            return r
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during solve: %s('%s')" % (str(e_type), str(value)), traceback

    def _eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError('unsat')

        return self._solver_backend.eval(e, n, extra_constraints=extra_constraints, result=self.result, solver=self._get_solver())

    def _batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError('unsat')

        return self._solver_backend.batch_eval(
                exprs,
                n,
                extra_constraints=extra_constraints,
                result=self.result,
                solver=self._get_solver()
        )

    def _max(self, e, extra_constraints=(), exact=None, cache=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError("Unsat during _max()")

        l.debug("Frontend.max() with %d extra_constraints", len(extra_constraints))

        two = self.eval(e, 2, extra_constraints=extra_constraints)
        if len(two) == 0: raise UnsatError("unsat during max()")
        elif len(two) == 1: return two[0]

        self.simplify()

        c = extra_constraints + (UGE(e, two[0]), UGE(e, two[1]))
        try:
            vals = self._solver_backend.max_values(e, extra_constraints=c, result=self.result, solver=self._get_solver())
            self._cache_eval(e, vals, exact=exact, cache=cache)
            return max(vals)
        except BackendError:
            pass

        try:
            return self._solver_backend.max(e, extra_constraints=c, result=self.result, solver=self._get_solver())
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during _max: %s('%s')" % (str(e_type), str(value)), traceback

    def _min(self, e, extra_constraints=(), exact=None, cache=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError("Unsat during _min()")

        l.debug("Frontend.min() with %d extra_constraints", len(extra_constraints))

        two = self.eval(e, 2, extra_constraints=extra_constraints)
        if len(two) == 0: raise UnsatError("unsat during min()")
        elif len(two) == 1: return two[0]

        self.simplify()

        c = extra_constraints + (ULE(e, two[0]), ULE(e, two[1]))
        try:
            vals = self._solver_backend.min_values(e, extra_constraints=c, result=self.result, solver=self._get_solver())
            self._cache_eval(e, vals, exact=exact, cache=cache)
            return min(vals)
        except BackendError:
            pass

        try:
            return self._solver_backend.min(e, extra_constraints=c, result=self.result, solver=self._get_solver())
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during _min: %s('%s')" % (str(e_type), str(value)), traceback

    def _solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        try:
            b = self._solver_backend.solution(e, v, extra_constraints=extra_constraints, solver=self._get_solver())
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during _solution: %s('%s')" % (str(e_type), str(value)), traceback

        return b

    def _is_true(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            b = self._solver_backend.is_true(e, extra_constraints=extra_constraints, solver=self._get_solver())
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during _is_true: %s('%s')" % (str(e_type), str(value)), traceback

        return b

    def _is_false(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            b = self._solver_backend.is_false(e, extra_constraints=extra_constraints, solver=self._get_solver())
        except BackendError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyFrontendError, "Backend error during _is_false: %s('%s')" % (str(e_type), str(value)), traceback

        return b

    #
    # Serialization and such.
    #

    def downsize(self):
        ConstrainedFrontend.downsize(self)
        self._tls.solver = None
        self._to_add = [ ]

    #
    # Merging and splitting
    #

    def _blank_copy(self):
        return FullFrontend(self._solver_backend, cache=self._cache, timeout=self.timeout)

    def branch(self):
        b = ConstrainedFrontend.branch(self)
        b._tls.solver = getattr(self._tls, 'solver', None) #pylint:disable=no-member
        b._to_add = list(self._to_add)
        return b

    def merge(self, others, merge_flag, merge_values):
        return self._solver_backend.__class__.__name__ == 'BackendZ3', ConstrainedFrontend.merge(self, others, merge_flag, merge_values)[1]

from ..errors import UnsatError, BackendError, ClaripyFrontendError
from ..ast.bv import UGE, ULE
from ..backend_manager import backends
